open Syntax
open Support.Error
open Format

(* Warning: doesn't do proper quoting, diamonds, etc *)

let string_unary o =
  match o.v with
  | NotOp     -> "not"
  | UMinusOp  -> "-"
  | UFminusOp -> "-." (* ? *)

let string_binary o =
  match o.v with
  | TimesOp   -> "*"
  | DivOp     -> "/"
  | PlusOp    -> "+"
  | MinusOp   -> "-"
  | FtimesOp  -> "*."
  | FdivOp    -> "/."
  | FplusOp   -> "+."
  | FminusOp  -> "-."
  | LessOp    -> "<"
  | LteqOp    -> "<="
  | GreaterOp -> ">"
  | GteqOp    -> ">="
  | EqualOp   -> "="
  | ConsOp    -> "cons"
  | AppendOp  -> "@" (* ? *)
  | AndOp     -> "&"
  | OrOp      -> "|"
  | AndalsoOp -> "andalso"
  | OrelseOp  -> "orelse"
  | ModOp     -> "%"


let rec string_value v =
  match v.v with
  | VarVal v    -> v
  | IntVal i    -> string_of_int i
  | FloatVal f  -> string_of_float f
  | CharVal c   -> Char.escaped c
  | StringVal s -> "\"" ^ s ^ "\""
  | BoolVal b   -> if b then "true" else "false"
  | UnitVal     -> "()"
  | UnaryOpVal (o,v) -> (string_unary o) ^ " (" ^ (string_value v) ^ ")"
  | BinaryOpVal (b,v1,v2) -> "(" ^ (string_value v1) ^ " " ^ (string_binary b) ^
                            " " ^ (string_value v2) ^ ")"
  | PairVal (v1,v2) -> "(" ^ (string_value v1) ^ "," ^ (string_value v2) ^ ")"

let rec print_expr e =
  open_box 2;
  print_string "(";
  begin
  match e.v with
  | ValueExp v  -> print_string (string_value v)
  | ConstrExp (c,vs,d) -> let d' = match d with
                                   | New -> ""
                                   | Reuse v -> "@" ^ v
                          in
                          if vs = []
                          then print_string (c ^ d')
                          else print_string (c ^ "(" ^ (String.concat "," (List.map string_value vs)) ^ ")" ^ d')
  | AppExp (f,vs,_) -> print_string (f ^ " " ^ (String.concat " " (List.map string_value vs)))
  | LetExp (v,e1,e2,_) -> print_string ("let " ^ v ^ " =");
                          print_space ();
                          print_expr e1;
                          print_space ();
                          print_string "in ";
                          print_expr e2
  | SeqExp (e1,e2) -> print_string "(";
                      print_expr e1;
                      print_string ");";
                      print_cut ();
                      print_string "(";
                      print_expr e2;
                      print_string ")"
  | IfExp (v,e1,e2) -> print_string ("if " ^ (string_value v));
                       print_space ();
                       print_string "then ";
                       print_expr e1;
                       print_space ();
                       print_string "else ";
                       print_expr e2
  | MatchPExp (v,v1,v2,e) -> print_string ("match " ^ v ^ " with (" ^ v1 ^ "," ^ v2 ^ ")");
                             print_space ();
                             print_expr e
  | MatchExp (v, rules) -> print_string ("match " ^ v ^ " with");
                           List.iter print_rule rules
  end;
  print_string ")";
  close_box ()

and print_rule (Matchrule (_,c,vs,_,e,d)) =
  print_space ();
  open_box 2;
  let d' = match d with
    | New -> "@_"
    | Reuse v -> "@" ^ v
    | Void -> ""
  in
  if vs = []
  then print_string ("| " ^ c ^ d' ^ " ->")
  else print_string ("| " ^ c ^ "(" ^ (String.concat "," vs) ^ ")" ^ d' ^ " ->");
  print_space ();
  print_expr e;
  close_box ()


let print_fun lets (FunctionDef (_,f,vs,e)) =
  open_box 2;
  print_string (lets ^ " " ^ f ^ " " ^ (String.concat " " vs) ^ " =");
  force_newline ();
  print_expr e;
  print_newline ()

let rec string_of_type ty =
  match ty.v with
    UnitTyp    -> "unit"
  | DiamondTyp -> "diamond"
  | BoolTyp    -> "bool"
  | IntTyp     -> "int"
  | FloatTyp   -> "float"
  | CharTyp    -> "char"
  | StringTyp  -> "string"
  | ConTyp(tys,name) -> name
  | ArrowTyp(t1,t2) -> (string_of_type t1) ^ " -> " ^ (string_of_type t2)
  | ArrayTyp t -> (string_of_type t) ^ " array"
  | PairTyp (t1,t2) -> (string_of_type t1) ^ " * " ^ (string_of_type t2)

let string_of_size s = 
  match s with
      Unspecified -> ""
    | Specified sz -> "(*" ^ (string_of_int sz) ^ "*)"

let print_tycon pre (TypCon (_,c,s,tys)) =
  let size = string_of_size s in
  match tys with
    [] -> print_string (pre ^ c ^ size)
  |  _ -> print_string (pre ^ c ^ size ^ " of " ^ (String.concat " * " (List.map string_of_type tys)))

let print_tydec (TypDec (_,_,id,tyc)) =
  print_string ("type " ^ id ^ " = ");
  (match tyc with [] -> ()
    | (h::t) -> print_tycon "" h;
        List.iter (print_tycon " | ") t);
  print_newline ()

let print_valdec v =
  match v with
    (ValDec (_,id,s,ty)) -> print_endline ("val " ^ id ^ (string_of_size s) ^
                                           " : " ^ string_of_type ty)
  | _ -> print_endline "(* unsupported valdec *)"

let print_program (Program (_,tys,vls,fs)) =
  List.iter print_tydec tys;
  print_newline ();
  List.iter print_valdec vls;
  print_newline ();
  List.iter (function [] -> ()
               | (h::t) -> print_fun "let" h;
                   List.iter (print_fun "and") t) fs
